//------------------------------------------------------------------------------------------------
/// @file EvaluatorGLO.cpp
/// @author Frolov Andrei
/// @brief     a EvaluatorGLO.
//------------------------------------------------------------------------------------------------

#include "GeomUtils.h"
#include "EvaluatorGLO.h"
#include "Mathematics/Rotation.h"

#include <iostream>
#include <sstream>
#include <algorithm>


/// @brief 
EvaluatorGLO::~EvaluatorGLO()
{
   if (m_file)
   {
      if (m_file->size)
         delete m_file->data;
      delete m_file;
   }
   ClearCachedData();
}

/// @brief       
/// @param[in] _numParameters -  
/// @param[in] _pParameters -   _numParameters 
INTEROP_RESULT EvaluatorGLO::SetInitParams(INTEROP_COUNT _numParameters, const FVDLL_PARAMETER* _pParameters)
{
   for (INTEROP_COUNT iParam = 0; iParam < _numParameters; ++iParam)
   {
      const FVDLL_PARAMETER& param = _pParameters[iParam];
      if (strcmp(param.id, "Filename") == 0)
      {
         if (!param.value.pFile->size)
         {
            fvapiLogMessage(FVAPI_MSG_ERROR, ERROR_IO_FILE);
            throw INTEROP_ERROR_INVALID_PARAMETER;
         }
         m_file = param.value.pFile;
      }
      else if ((strcmp(param.id, "IsExtapolationOn") == 0) && (param.type == FVDLL_PARAM_BOOLEAN))
      {
         m_IsExtrapolationOn = param.value.boolean;
      }
      else if ((strcmp(param.id, "DefaultValue") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_defaultValue = param.value.real;
      }
      else if ((strcmp(param.id, "Translation/X") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_translation[0] = param.value.real;
      }
      else if ((strcmp(param.id, "Translation/Y") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_translation[1] = param.value.real;
      }
      else if ((strcmp(param.id, "Translation/Z") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_translation[2] = param.value.real;
      }
      else if ((strcmp(param.id, "Rotation/Angle") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_angle = param.value.real;
      }
      else if ((strcmp(param.id, "Rotation/RotationVector/X") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_rotationVector[0] = param.value.real;
      }
      else if ((strcmp(param.id, "Rotation/RotationVector/Y") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_rotationVector[1] = param.value.real;
      }
      else if ((strcmp(param.id, "Rotation/RotationVector/Z") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_rotationVector[2] = param.value.real;
      }
      else if ((strcmp(param.id, "Tolerance") == 0) && (param.type == FVDLL_PARAM_REAL))
      {
         m_tolerance = std::max(param.value.real, EPS);
         m_tolerance2 = m_tolerance * m_tolerance;
      }
      else
      {
         throw INTEROP_ERROR_INVALID_PARAMETER;
      }
   }
   ReadValues();                             //    
   TransformCoordinates();                   //  
   GetBoundingBox();                         //  
   m_pointGrid.Initialize(m_points, m_bBox); //  m_pointGrid;
   return INTEROP_OK;
}

/// @brief      
/// @param[in] _hVarContext -   
/// @return    
INTEROP_DOUBLE EvaluatorGLO::EvaluateScalar(FVAPI_CONTEXT _hVarContext)
{
   INTEROP_UINT32 contentMask;
   double calcPoint[3];

   fvapiGetContextMask(_hVarContext, &contentMask);

   //   ,     ->    
   if (contentMask & FVAPI_VCTX_CELL &&
       contentMask & FVAPI_VCTX_CHIP) 
   {
      FVAPI_FACE face;
      FVAPI_CELL cell;
      INTEROP_UINT32 iHCell;
      INTEROP_UINT32 iCell;

      fvapiGetFaceOfContext(_hVarContext, &face);
      fvapiGetCellOfContext(_hVarContext, &cell);
      fvapiGetFaceCenter(face, cell, calcPoint);
      fvapiGetCellGlobalIndex(cell, &iHCell, &iCell);

      //      
      if (iHCell >= m_HCellCount)
      {
         std::lock_guard<std::mutex> locker(m_mutex);

         m_HCellFaces.reserve(1 + (size_t)iHCell);
         m_usedFaces.reserve(1 + (size_t)iHCell);
         for (; m_HCellCount <= iHCell; ++m_HCellCount)
         {
            m_usedFaces.push_back(new std::set<FVAPI_FACE>);
            m_HCellFaces.push_back(new faceMap);
         }
      }
      auto& usedFaces = *m_usedFaces[iHCell];
      auto& HCellFaces = *m_HCellFaces[iHCell];
      usedFaces.insert(face);
      if (HCellFaces.find(face) == HCellFaces.end())
      {
         HCellFaces[face] = GetValueInPoint(calcPoint); 
      }
      return HCellFaces[face];
   }
   else if (contentMask & FVAPI_VCTX_POINT) //      ,    
   {
      fvapiGetPointOfContext(_hVarContext, calcPoint);
      return GetValueInPoint(calcPoint); //         ( )
   }
   else
   {
      return FLT_MAX;
   }
}

/// @brief     .
/// @details        . 
INTEROP_RESULT EvaluatorGLO::BeforeTimeStep()
{
   for (size_t i = 0; i < m_usedFaces.size(); ++i)
   {
      m_usedFaces[i]->clear();
   }
   return INTEROP_OK;
}

/// @brief     
/// @details      .    
INTEROP_RESULT EvaluatorGLO::AfterTimeStep()
{
   INTEROP_UINT32 HCellCount;
   INTEROP_DOUBLE curTime;
   fvapiGetAllHyperCellCount(&HCellCount);
   fvapiGetCurrentTime(&curTime);

   //      
   if (curTime == 0)
   {
      ClearCachedData();
   }

   //      
   if (m_HCellCount > HCellCount + 1)
   {
      for (size_t i = m_HCellCount - 1; i > HCellCount; --i)
      {
         m_usedFaces[i]->clear();
         delete m_usedFaces[i];
         m_HCellFaces[i]->clear();
         delete m_HCellFaces[i];
      }
      m_HCellCount = HCellCount + 1;
      m_HCellFaces.resize(m_HCellCount);
      m_usedFaces.resize(m_HCellCount);
   }

   //        
   for (INTEROP_UINT32 i = 0; i < m_HCellCount; ++i)
   {
      auto& usedFaces = *m_usedFaces[i];
      auto& HCellFaces = *m_HCellFaces[i];
      if (usedFaces.size() != HCellFaces.size()) // -   /
      {
         std::set<FVAPI_FACE> toDelete;
         for (std::pair<FVAPI_FACE, double> faceID : HCellFaces)
         {
            if (usedFaces.find(faceID.first) == usedFaces.end()) //      
            {
               toDelete.insert(faceID.first);
            }
         }
         for (FVAPI_FACE key : toDelete)
         {
            HCellFaces.erase(key);
         }
         toDelete.clear();
      }
   }
   return INTEROP_OK;
}

/// @brief     ;
void EvaluatorGLO::ClearCachedData()
{
   for (size_t i = 0; i < m_HCellFaces.size(); ++i)
      m_HCellFaces[i]->clear();
   for (size_t i = 0; i < m_usedFaces.size(); ++i)
      m_usedFaces[i]->clear();
}

/// @brief      m_file   m_points  m_polygons
void EvaluatorGLO::ReadValues()
{
   std::istringstream file(reinterpret_cast<char*>(m_file->data)); 
   if (!file || file.bad())
   {
      fvapiLogMessage(FVAPI_MSG_ERROR, ERROR_PARAMETER_FILE);
      throw INTEROP_ERROR_INVALID_PARAMETER;
   }

   std::string line;
   static const std::string NumPoints = "NumPoints";
   bool isNumPointsFound = false;

   try
   {   
      //    (,       GLO    ).    
      while (getline(file, line))
      {
         if (line.size() == 0 || line[0] == '#')
         {
            continue;
         }
         else if (line.find(NumPoints) != std::string::npos)
         {
            isNumPointsFound = true;
            break; /// <     
         }
      }

      //    
      if (isNumPointsFound)
      {
         //  
         int pointCount = stoi(line.substr(line.find_first_of('=') + 1));
         getline(file, line); /// <   :    # -------------------

         //  
         double x, y, z, v;
         m_rawCoordinates.reserve(pointCount);
         m_rawValues.reserve(pointCount);
         m_points.reserve(pointCount);
         for (int i = 0; i < pointCount && getline(file, line); ++i)
         {
            sscanf_s(line.c_str(), "%lg %lg %lg %lg", &x, &y, &z, &v);
            m_rawCoordinates.push_back(Vector3{ x, y, z });
            m_rawValues.push_back(v);
            m_points.push_back(ValuePoint(&m_rawCoordinates.back(), m_rawValues.back()));
         }
      }
      //    ,   , ,  .     
      else
      {
         //   (,     )
         file.seekg(std::istringstream::beg);
         double x, y, z, v;
         do
         {
            sscanf_s(line.c_str(), "%lg %lg %lg %lg", &x, &y, &z, &v);
            m_rawCoordinates.push_back(Vector3{x, y, z});
            m_rawValues.push_back(v);
            m_points.push_back( ValuePoint(&m_rawCoordinates.back(), m_rawValues.back()) );
         } while (getline(file, line));
      }

      //      
      if (m_points.empty())
      {
         throw;
      }
   }
   catch (...)
   {
      fvapiLogMessage(FVAPI_MSG_ERROR, ERROR_PARAMETER_FILE);
      throw INTEROP_ERROR_INVALID_PARAMETER;
   }
}

/// @brief        
void EvaluatorGLO::TransformCoordinates()
{
   if (m_angle != 0.0)
   {
      gte::AxisAngle<3, double> rotationPair(m_rotationVector, m_angle);
      gte::Normalize(rotationPair.axis); //     
      gte::Rotation<3, double> rotation(rotationPair);
      gte::Matrix<3, 3, double> rotationMatrix = rotation;
      for (auto& point : m_rawCoordinates)
      {
         point = rotationMatrix * point;
      }
   }
   if (m_translation != Vector3::Zero())
   {
      for (auto& point : m_rawCoordinates)
      {
         point += m_translation;
      }
   }
}

/// @brief    
void EvaluatorGLO::GetBoundingBox()
{
   gte::GetContainer(static_cast<int>(m_rawCoordinates.size()), m_rawCoordinates.data(), m_bBox);

   //     
   m_bBoxTolerance = m_bBox;
   m_bBoxTolerance.min[0] -= m_tolerance;
   m_bBoxTolerance.min[1] -= m_tolerance;
   m_bBoxTolerance.min[2] -= m_tolerance;
   m_bBoxTolerance.max[0] += m_tolerance;
   m_bBoxTolerance.max[1] += m_tolerance;
   m_bBoxTolerance.max[2] += m_tolerance;
}

/// @brief         
/// @param _calcPoint -  
/// @return   
double EvaluatorGLO::GetValueInPoint(double* _calcPoint) const
{
   Vector3 point{_calcPoint[0], _calcPoint[1], _calcPoint[2]};
   if (gte::InContainer(point, m_bBoxTolerance)) //     
   {
      return GetPointInterpolatedValue(point);
   }
   else if (m_IsExtrapolationOn)
   {
      return GetPointExtrapolatedValue(point); //     
   }
   return m_defaultValue;
}

/// @brief     
/// @param _pnt   
/// @return     . 
/// @detail ,     bBox.    ,  
double EvaluatorGLO::GetPointInterpolatedValue(Vector3& _pnt) const
{
   Vector3 position = (_pnt - m_bBox.min) / m_pointGrid.splitStep;
   int index_x = GetClippedIndex(position[0], m_pointGrid.maxInd.x);
   int index_y = GetClippedIndex(position[1], m_pointGrid.maxInd.y);
   int index_z = GetClippedIndex(position[2], m_pointGrid.maxInd.z);

   //       ,       ,
   //       
   //   ,       
   bool isGoodBox = true;
   if (m_pointGrid.At(index_x, index_y, index_z)->size() == 0)
   {
      isGoodBox = false;
   }

   sValBox valBox(_pnt);
   int iteration = 0;
   static const int max_iteration = 4;

   //    0  1   27  ( 333)
   FillValBox(valBox, index_x, index_y, index_z, iteration++);
   FillValBox(valBox, index_x, index_y, index_z, iteration++);
   
   //       ,     
   while (iteration <= max_iteration && !(valBox.IsValid()))
   {
      FillValBox(valBox, index_x, index_y, index_z, iteration++);
   }
   if (isGoodBox && valBox.IsValid())
      return valBox.GetInterpolatedValue();
   else
   {
      if (valBox.IsInTolerance(m_tolerance)) //   
      {
         return valBox.GetInterpolatedValue();
      }
      else if (m_IsExtrapolationOn)
      {
         return GetPointExtrapolatedValue(_pnt); //   
      }
   }
   return m_defaultValue;
}

/// @brief            
/// @param _pnt -  
/// @return  -   ,   .
double EvaluatorGLO::GetPointExtrapolatedValue(Vector3& _pnt) const
{
   //  
   double valStorage = 0;
   double distStorage = 0;
   double dist2;
   for (auto& point : m_points)
   {
      dist2 = std::max(Length2(_pnt - *point.coord), DBL_MIN);
      valStorage += point.m_value / dist2;
      distStorage += 1. / dist2;
   }

   return valStorage / distStorage;
}

/// @brief    ,     
/// @param _box 
/// @param _x @param _y @param _z -      m_pointGrid
/// @param _iter -   
void EvaluatorGLO::FillValBox(sValBox& _box, int _x, int _y, int _z, int _iter) const
{
   if (_iter == 0)
   {
      auto container = m_pointGrid.At(_x, _y, _z);
      for (auto point : *container)
      {
         _box.Insert(point);
      }
      return;
   }

   int x_min = std::max(0, _x - _iter);
   int y_min = std::max(0, _y - _iter);
   int z_min = std::max(0, _z - _iter);

   int x_max = std::min<int>(m_pointGrid.maxInd.x, _x + _iter);
   int y_max = std::min<int>(m_pointGrid.maxInd.y, _y + _iter);
   int z_max = std::min<int>(m_pointGrid.maxInd.z, _z + _iter);

   int inner_threshold = _iter - 1;

   for (int x = x_min; x <= x_max; x++)
   {
      for (int y = y_min; y <= y_max; y++)
      {
         for (int z = z_min; z <= z_max; z++)
         {
            //     
            if ((abs(x - _x) > inner_threshold ||
                 abs(y - _y) > inner_threshold ||
                 abs(z - _z) > inner_threshold))
            {
               auto container = m_pointGrid.At(x, y, z);
               for (auto point : *container)
               {
                  _box.Insert(point);
               }
            }
         }
      }
   }
}
