#include "GeomUtils.h"

#include <Mathematics/DistPointLine.h>
#include <Mathematics/DistPointTriangle.h>

#include <array>

//========================================================================================
//     
//========================================================================================

static constexpr char lineCombinationsAmount[] = 
{
   0, 1, 3, 6, 10, 15
};

static constexpr char lineLookupTable[] =
{
   0, 1,
   0, 2,   1, 2,
   0, 3,   1, 3,   2, 3,
   0, 4,   1, 4,   2, 4,   3, 4,
   0, 5,   1, 5,   2, 5,   3, 5,   4, 5
};

static constexpr char triangleCombinationsAmount[] =
{
   0, 0, 1, 4, 10, 20
};

static constexpr char triangleLookupTable[] =
{
   0, 1, 2,
   0, 1, 3,   0, 2, 3,   1, 2, 3,
   0, 1, 4,   0, 2, 4,   0, 3, 4,   1, 2, 4,   1, 3, 4,   2, 3, 4,
   0, 1, 5,   0, 2, 5,   0, 3, 5,   0, 4, 5,   1, 2, 5,   1, 3, 5,   1, 4, 5,   2, 3, 5,   2, 4, 5,   3, 4, 5
};

//========================================================================================
// class PointGrid -   ,    
//========================================================================================
/// @brief .   gridBoxes
PointGrid::~PointGrid()
{
   for (auto box : gridBoxes)
      delete box;
   gridBoxes.clear();
}

/// @brief      
/// @param _points -   
/// @param _bBox   -   
void PointGrid::Initialize(std::vector<ValuePoint>& _points, gte::AlignedBox3<double>& _bBox)
{
   //  
   Vector3 spaces = _bBox.max - _bBox.min;
   double max_space = std::max(spaces[0], std::max(spaces[1], spaces[2]));

   //     
   double unit_box_point_count = pow(_points.size(), 0.33); //         
   splitStep = max_space / unit_box_point_count;

   //      (    ,      1)
   size.x = std::max<int>(static_cast<int>(spaces[0] / splitStep), 1);
   size.y = std::max<int>(static_cast<int>(spaces[1] / splitStep), 1);
   size.z = std::max<int>(static_cast<int>(spaces[2] / splitStep), 1);
   x_shift = size.y * size.z;

   int pointGridSize = size.x * size.y * size.z;
   gridBoxes.reserve(static_cast<size_t>(pointGridSize + 1));
   for (int i = 0; i < pointGridSize; ++i)
      gridBoxes.push_back(new PointContainer());

   // 
   maxInd.x = size.x - 1;
   maxInd.y = size.y - 1;
   maxInd.z = size.z - 1;
   int x_ind, y_ind, z_ind;
   Vector3 relCoord;
   for (ValuePoint& pnt : _points)
   {
      relCoord = *pnt.coord - _bBox.min;
      x_ind = GetClippedIndex(relCoord[0] / splitStep, maxInd.x);
      y_ind = GetClippedIndex(relCoord[1] / splitStep, maxInd.y);
      z_ind = GetClippedIndex(relCoord[2] / splitStep, maxInd.z);
      At(x_ind, y_ind, z_ind)->push_back(&pnt);
   }
}

/// @brief     
/// @param x @param y @param z -  
/// @return       
PointContainer* PointGrid::At(size_t x, size_t y, size_t z) const
{
   return gridBoxes.at(x * x_shift + y * size.z + z);
}


//========================================================================================
// class sValBox - bBox      -  
//========================================================================================

/// @brief 
/// @param _center -  
sValBox::sValBox(Vector3& _center)
{
   distances = { DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX, DBL_MAX };
   points.resize(6, nullptr);
   center = _center;
}

/// @brief   ,          
bool sValBox::IsValid() const
{
   return std::none_of(points.begin(), points.end(), [](auto pntIter) { return pntIter == nullptr; });
}

/// @brief   ,         
bool sValBox::IsEmpty() const
{
   return std::all_of(points.begin(), points.end(), [](auto pntIter) { return pntIter == nullptr; });
}

/// @brief    sValBox,     
/// @param _pnt -   
void sValBox::Insert(ValuePoint* _pnt)
{
   double dist = Length2(*_pnt->coord - center);
   double cur_point_direction_coord;
   double calc_point_direction_coord;
   double center_point_direction_coord;
   int direction;
   //  
   for (int dir = 0; dir < 6; dir += 2)
   {
      if (dist < distances[dir])
      {
         direction = dir / 2;
         cur_point_direction_coord = points[dir] ? (*points[dir]->coord)[direction] : -DBL_MAX;
         calc_point_direction_coord = (*_pnt->coord)[direction];
         center_point_direction_coord = center[direction] + EPS;

         if (cur_point_direction_coord <= calc_point_direction_coord &&
             calc_point_direction_coord <= center_point_direction_coord)
         {
            distances[dir] = dist;
            points[dir] = _pnt;
         }
      }
   }

   //  
   for (int dir = 1; dir < 6; dir += 2)
   {
      if (dist < distances[dir])
      {
         direction = dir / 2;
         cur_point_direction_coord = points[dir] ? (*points[dir]->coord)[direction] : DBL_MAX;
         calc_point_direction_coord = (*_pnt->coord)[direction];
         center_point_direction_coord = center[direction] - EPS;

         if (cur_point_direction_coord >= calc_point_direction_coord &&
             calc_point_direction_coord >= center_point_direction_coord)
         {
            distances[dir] = dist;
            points[dir] = _pnt;
         }
      }
   }
}

/// @brief      
/// @return     . FLT_MAX,    
double sValBox::GetInterpolatedValue() const
{
   //   ,   
   std::set<ValuePoint*> closest_points;
   for (auto pnt : points)
      if (pnt)
         closest_points.insert(pnt);
   if (closest_points.size() == 0)
      return FLT_MAX;
   if (closest_points.size() == 1)
      return (*closest_points.begin())->m_value; //      

   //  
   double valStorage = 0;
   double distStorage = 0;
   double dist2;
   for (auto point : closest_points)
   {
      dist2 = std::max(Length2(center - *point->coord), DBL_MIN);
      valStorage += point->m_value / dist2;
      distStorage += 1. / dist2;
   }

   return valStorage / distStorage;
}

/// @brief   ,  valBox-    
/// @param _tolerance -      
/// @return       
bool sValBox::IsInTolerance(double _tolerance) const
{
   //   ,   
   std::set<ValuePoint*> closest_points;
   for (auto pnt : points)
      if (pnt)
         closest_points.insert(pnt);

   //  -    ,   true;
   for (auto pnt : closest_points)
   {
      if (Length2(center - *pnt->coord) <= (_tolerance * _tolerance))
         return true;
   }

   size_t size = closest_points.size();
   switch (size)
   {
   case 0:
   case 1:
      return false; //   ,       ,  false
   case 2:
      return CheckLines(closest_points, _tolerance);
   default:
      return CheckLines(closest_points, _tolerance) || CheckPolygons(closest_points, _tolerance);
   }
   return false; 
}

/// @brief        _points
/// @param _points -  
/// @param _tolerance -       
/// @return       
/// @remark     
bool sValBox::CheckLines(std::set<ValuePoint*>& _points, double _tolerance) const
{
   std::vector<Vector3*> pnts;
   for (auto point : _points)
   {
      pnts.push_back(point->coord);
   }

   char combAmount = lineCombinationsAmount[pnts.size()];
   gte::Line3<double> line;
   gte::DCPPoint3Line3<double> point_line_distance;

   for (char index = 0; index < combAmount; ++index)
   {
      line.origin = *pnts[lineLookupTable[2 * index]];
      line.direction = line.origin - *pnts[lineLookupTable[2 * index + 1]];
      gte::Normalize(line.direction); //      
      if (point_line_distance(center, line).distance <= _tolerance)
         return true;
   }
   return false;
}

/// @brief       _points
/// @param _points -  
/// @param _tolerance -       
/// @return      
bool sValBox::CheckPolygons(std::set<ValuePoint*>& _points, double _tolerance) const
{
   std::vector<Vector3*> pnts;
   for (auto point : _points)
   {
      pnts.push_back(point->coord);
   }

   char combAmount = triangleCombinationsAmount[pnts.size()];
   gte::Triangle3<double> triangle;
   gte::DCPPoint3Triangle3<double> point_triangle_distance;

   for (char index = 0; index < combAmount; ++index)
   {
      triangle = { *pnts[lineLookupTable[3 * index]],
                   *pnts[lineLookupTable[3 * index + 1]],
                   *pnts[lineLookupTable[3 * index + 2]],
      }; 
      if (point_triangle_distance(center, triangle).distance <= _tolerance)
         return true;
   }
   return false;
}
